/* ez-blob-tagz.c: A tagged blob serialiser.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

/*
  Tagged blob encoding.

  This is a streamable self-describing byte-oriented binary format.
  It is a general purpose format and supports a super-set of the
  ez_blob descriptor.  It supports primitive and struct types and
  sequences thereof and there is room for extension.

  Each item beings with a descriptor byte, then followed by a tag id,
  a possible count, and the payload.

 xxxxttcc control byte

  xxxx type code

   0 uint8       unsigned int, value zero-extended
   1 uint16
   2 uint32
   3 uint64
   - reserved
   5 float16
   6 float32
   7 float64
   - reserved
   f struct

  note that for int/float types, code&3 == log2 of element size in bytes

  tt log2 of tag size in bytes

   0 1 byte
   1 2 byte
   2 4 byte
   3 reserved, no tag?

  cc log2 of count size in bytes, used to indicate sequence length or non-sequence.

   0 1 byte
   1 2 byte
   2 4 byte
   3 none, single item follows

   ff is struct-end code

 A header is a control byte followed by an optional 1/2/4 byte-length tag,
 followed by an optional 1/2/4 byte-length count.

 A structure payload is a list of tagged fields until a struct-end
 code.  A structure sequence is a list of count struct-encoded blocks.

 Integers can be stored in the smallest number of bytes, i.e. with
 all leading $00 bytes removed.

 Possible extension is signed types and ignoring leading $ff bytes.

ez_blob requirements:

  All fields must have an id.
  The id for each field must be unique for any given struct.
  The ids must be in incremental order.
  Arrays must be stored at their natural size.

final note:

  This is a bit over-engineered.

*/

#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "ez-list.h"

#include "ez-blob.h"
#include "ez-blob-tagz.h"
#include "ez-blob-io.h"

//#define D(x) do { x; fflush(stdout); } while (0)
#define D(x)

#define EZT_INT8    0x00
#define EZT_INT16   0x10
#define EZT_INT32   0x20
#define EZT_INT64   0x30

#define EZT_FLOAT16 0x50
#define EZT_FLOAT32 0x60
#define EZT_FLOAT64 0x70
#define EZT_STRUCT  0xf0

#define EZT_TYPE    0xf0

#define EZT_COUNTSHIFT 0
#define EZT_TAGSHIFT   2
#define EZT_DATASHIFT  4

#define EZT_COUNTSIZE  (0x3 << EZT_COUNTSHIFT)
#define EZT_TAGSIZE    (0x3 << EZT_TAGSHIFT)
#define EZT_DATASIZE   (0x3 << EZT_DATASHIFT)

#define EZT_NOCOUNT (0x3 << EZT_COUNTSHIFT)

#define EZT_END    0xff // well, end-struct

static void blobio_writei(struct ez_blobio *io, int sc, uint64_t val) {
	switch (sc) {
	case 0:
		blobio_write8(io, val);
		break;
	case 1:
		blobio_write16(io, val);
		break;
	case 2:
		blobio_write32(io, val);
		break;
	case 3:
		blobio_write64(io, val);
		break;
	}
}

// log2 of significant bytes in value
static int size_code(uint64_t size) {
	if (size <= 0xff)
		return 0x0;
	else if (size <= 0xffff)
		return 0x1;
	else if (size <= 0xffffffff)
		return 0x2;
	else
		return 0x3;	
}

static int count_code(uint32_t size) {
	if (size <= 0xff)
		return 0x0;
	else if (size <= 0xffff)
		return 0x1;
	else
		return 0x2;
}

static void blobio_writet(struct ez_blobio *io, uint8_t type, uint32_t tag) {
	int tc = size_code(tag);

	blobio_writeb(io, type | (tc << EZT_TAGSHIFT));
	blobio_writei(io, tc, tag);
}

static void tagz_encode_struct(struct ez_blobio *io, int id, const ez_blob_desc *desc, const void *p) {
	blobio_writet(io, EZT_STRUCT | EZT_NOCOUNT, id);
	
	for (int i=0,dlen=desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];
		const void *v = (p + d->bd_offset);
		uint64_t val = 0;
		uint32_t count = 0;
		const void *ptr = NULL;
		
		switch (d->bd_type & EZ_BLOB_TYPE) {
		case EZ_BLOB_INT8:
			val = *(uint8_t *)v;
			goto doint;
		case EZ_BLOB_INT16:
			val = *(uint16_t *)v;
			goto doint;
		case EZ_BLOB_INT32:
			val = *(uint32_t *)v;
			goto doint;
		case EZ_BLOB_INT64:
			val = *(uint64_t *)v;
		doint:
			if (val) {
				int sc = size_code(val);
				
				blobio_writet(io, EZT_INT8 | EZT_NOCOUNT | (sc << EZT_DATASHIFT), d->bd_id);
				blobio_writei(io, sc, val);
			}
			break;
		case EZ_BLOB_FLOAT32:
			blobio_writet(io, EZT_FLOAT32 | EZT_NOCOUNT, d->bd_id);
			blobio_writef(io, *(const float *)v);
			break;
		case EZ_BLOB_FLOAT64:
			blobio_writet(io, EZT_FLOAT64 | EZT_NOCOUNT, d->bd_id);
			blobio_writed(io, *(const double *)v);
			break;
		case EZ_BLOB_STRING:
			// NULL strings are the default, so implement same but allow for empty strings
			ptr = *(const void **)v;
			if (ptr) {
				count = strlen(ptr);
				goto doblob;
			}
			break;
		case EZ_BLOB_BLOB:
			// empty blobs are the default, so implement same, empty are not encoded
			count = ((ez_blob *)v)->size;
			if (count) {
				int cc;
				
				ptr = ((ez_blob *)v)->data;
			doblob:
				cc = count_code(count);
				blobio_writet(io, EZT_INT8 | (cc << EZT_COUNTSHIFT), d->bd_id);
				blobio_writei(io, cc, count);
				blobio_write(io, ptr, count);
			}
			break;
		case EZ_BLOB_STRUCT:
			tagz_encode_struct(io, d->bd_id, d->bd_table, v);
			break;
		case EZ_BLOB_POINTER:
			ptr = *(const void **)v;
			if (ptr)
				tagz_encode_struct(io, d->bd_id, d->bd_table, ptr);
			break;
		case EZ_BLOB_LIST:
			count = ez_list_size((ez_list *)v);
			if (count) {
				int cc = count_code(count);
				
				blobio_writet(io, EZT_STRUCT | (cc << EZT_COUNTSHIFT), d->bd_id);
				blobio_writei(io, cc, count);
			
				for (ez_node *w = ez_list_head((ez_list *)v), *n = ez_node_succ(w);
				     n;
				     w = n, n=ez_node_succ(n)) {
					// which id to use?
					tagz_encode_struct(io, d->bd_table[0].bd_id, d->bd_table, w);
				}
			}
			break;
		}

	}

	blobio_writeb(io, EZT_END);
}

size_t ez_tagz_size(const ez_blob_desc *d, const void *p) {
	struct ez_blobio io = {
		.mode = BLOBIO_WRITE_SIZE
	};
	tagz_encode_struct(&io, d->bd_id, d, p);

	return io.index;
}

int ez_tagz_encode_raw(const ez_blob_desc *d, const void *p, ez_blob *blob) {
	struct ez_blobio io = {
		.data = blob->data,
		.size = blob->size,
		.mode = BLOBIO_WRITE_FIXED
	};

	tagz_encode_struct(&io, d->bd_id, d, p);

	return io.error;
}

int ez_tagz_encode(const ez_blob_desc *d, const void *p, ez_blob *blob) {
	struct ez_blobio io = {
		.mode = BLOBIO_WRITE_ALLOC
	};

	tagz_encode_struct(&io, d->bd_id, d, p);

	if (!io.error) {
		blob->size = io.index;
		blob->data = io.data;
		return 0;
	}
	free(blob->data);
	blob->data = 0;
	blob->size = 0;
	
	return -1;
}

static int blobio_readb(struct ez_blobio *io) {
	uint8_t *v = blobio_take(io, 1);
	return v ? *v : EZT_END;
}

static uint64_t blobio_readi(struct ez_blobio *io, int sc) {
	switch (sc) {
	case 0:
		return blobio_readi8(io);
	case 1:
		return blobio_readi16(io);
	case 2:
		return blobio_readi32(io);
	case 3:
		return blobio_readi64(io);
	}
	return 0;
}

static void tagz_decode_struct(struct ez_blobio *io, const ez_blob_desc *desc, void *p);
static void tagz_decode_fields(struct ez_blobio *io, const ez_blob_desc *desc, void *p);

static void tagz_decode_fields(struct ez_blobio *io, const ez_blob_desc *desc, void *p) {
	uint8_t h;
	const ez_blob_desc *dscan = desc ? desc + 1 : desc;
	const ez_blob_desc *dend = desc ? desc + desc->bd_length + 1 : desc;

	while ((h = blobio_readb(io)) != EZT_END) {
		uint32_t ftag;
		uint32_t fcount = 0;
		int sc = (h & EZT_DATASIZE) >> EZT_DATASHIFT; /* for primitives */
		int use_count = 0;

		ftag = blobio_readi(io, (h & EZT_TAGSIZE) >> EZT_TAGSHIFT);
		if ((h & EZT_COUNTSIZE) != EZT_NOCOUNT) {
			fcount = blobio_readi(io, (h & EZT_COUNTSIZE) >> EZT_COUNTSHIFT);
			use_count = 1;
		}

		// This forces the tags to be in order
		// It ensures each field can only be visited once
		const ez_blob_desc *d = NULL;
		{
			while (dscan < dend && dscan->bd_id < ftag)
				dscan++;
			if (dscan < dend && dscan->bd_id == ftag)
				d = dscan++;
		}

		if (p && d) {
			void *v = p + d->bd_offset;

			switch (h & EZT_TYPE) {
			case EZT_INT8:
			case EZT_INT16:
			case EZT_INT32:
			case EZT_INT64:
				if (use_count) {
					char *mem;
					size_t len = fcount * (1<<sc);
					void *data = blobio_take(io, len);

					if (!data)
						goto error;

					switch (d->bd_type & EZ_BLOB_TYPE) {
					case EZ_BLOB_STRING:
						mem = malloc(len+1);
						if (!mem)
							goto error;
						*(char **)v = mem;
						memcpy(mem, data, len);
						mem[len] = 0;
						//*(char **)v = data;
						D(printf("string %d '%s'\n", d->bd_id, mem));
						break;
					case EZ_BLOB_BLOB:
						mem = malloc(len);
						if (!mem)
							goto error;
						((ez_blob *)v)->data = mem;
						((ez_blob *)v)->size = len;
						memcpy(mem, data, len);
						//((ez_blob *)v)->data = data;
						break;
					}
				} else {
					switch (d->bd_type) {
					case EZ_BLOB_INT8:
						*(uint8_t *)v = blobio_readi(io, sc);
						break;
					case EZ_BLOB_INT16:
						*(uint16_t *)v = blobio_readi(io, sc);
						break;
					case EZ_BLOB_INT32:
						*(uint32_t *)v = blobio_readi(io, sc);
						break;
					case EZ_BLOB_INT64:
						*(uint64_t *)v = blobio_readi(io, sc);
						break;
					default:
						goto error;
					}
				}
				break;
			case EZT_FLOAT32:
				if (use_count) {
					blobio_take(io, fcount * 4);
				} else {
					switch (d->bd_type) {
					case EZ_BLOB_FLOAT32:
						*(float *)v = blobio_readf(io);
						break;
					case EZ_BLOB_FLOAT64:
						*(double *)v = blobio_readf(io);
						break;
					default:
						goto error;
					}
				}
				break;
			case EZT_FLOAT64:
				if (use_count) {
					blobio_take(io, fcount * 8);
				} else {
					switch (d->bd_type) {
					case EZ_BLOB_FLOAT32:
						*(float *)v = blobio_readd(io);
						break;
					case EZ_BLOB_FLOAT64:
						*(double *)v = blobio_readd(io);
						break;
					default:
						goto error;
					}
				}
				break;
			case EZT_STRUCT:
				if (use_count) {
					switch (d->bd_type) {
					case EZ_BLOB_LIST:
						for (int i=0; !io->error && i < fcount; i++) {
							void *node = ez_blob_alloc(d->bd_table);

							if (!node)
								goto error;

							tagz_decode_struct(io, d->bd_table, node);
							ez_list_addtail((ez_list *)v, node);
						}
						break;
					default:
						goto error;
					}
				} else {
					switch (d->bd_type & EZ_BLOB_TYPE) {
					case EZ_BLOB_STRUCT:
						D(printf("struct %d\n", d->bd_id));
						//ez_blob_init(d->bd_table, v);
						tagz_decode_fields(io, d->bd_table, v);
						break;
					case EZ_BLOB_POINTER:
						D(printf("pointer %d\n", d->bd_id));
						*(void **)v = ez_blob_alloc(d->bd_table);

						if (!*(void **)v)
							goto error;
						
						tagz_decode_fields(io, d->bd_table, *(void **)v);
						break;
					default:
						goto error;
					}
				}
				break;
			}
		} else {
			// Just skip fields
			switch (h & EZT_TYPE) {
			case EZT_INT8:
			case EZT_INT16:
			case EZT_INT32:
			case EZT_INT64:
			case EZT_FLOAT32:
			case EZT_FLOAT64:
				if (use_count) {
					blobio_take(io, fcount * (1<<sc));
				} else {
					blobio_take(io, (1<<sc));
				}
				break;
			case EZT_STRUCT:
				if (fcount) {
					for (int i=0; !io->error && i < fcount; i++) {
						tagz_decode_struct(io, NULL, NULL);
					}
				} else {
					tagz_decode_fields(io, NULL, NULL);
				}
				break;
			default:
				goto error;
			}
		}
	}
	return;
 error:
	io->index = io->size;
	io->error = 1;
}

static void tagz_decode_struct(struct ez_blobio *io, const ez_blob_desc *desc, void *p) {
	uint8_t h = blobio_readb(io);
	uint32_t stag;
	//uint32_t scount = 1;
		
	// must be a struct
	// what about a struct list? hmm
	if (io->error || (h & EZT_TYPE) != EZT_STRUCT) {
		//io->error = 1;
		return;
	}

	stag = blobio_readi(io, (h & EZT_TAGSIZE) >> EZT_TAGSHIFT);
	stag = stag;
	//if (stag != desc->bd_id) {
		// ??
	//}
	//if (h & EZT_COUNTSIZE)
	//	count = blobio_readi(io, (h & EZT_COUNTSIZE) >> EZT_COUNTSHIFT);

	// read fields
	tagz_decode_fields(io, desc, p);
}

int ez_tagz_decode_raw(const ez_blob_desc *desc, const ez_blob *blob, void *p) {
	struct ez_blobio io = {
		.size = blob->size,
		.data = blob->data,
		.mode = BLOBIO_READ
	};

	ez_blob_init(desc, p);
	tagz_decode_struct(&io, desc, p);

	if (io.error) {
		ez_blob_free_raw(desc, p);
		return -1;
	}

	return 0;
}

void *ez_tagz_decode(const ez_blob_desc *desc, const ez_blob *blob) {
	void *p = malloc(desc->bd_offset);

	if (p && ez_tagz_decode_raw(desc, blob, p) != 0) {
		free(p);
		p = NULL;
	}

	return p;
}

/* ********************************************************************** */
// TODO: move to separate .o
static void tagz_dump_struct(struct ez_blobio *io, int depth);
static void tagz_dump_fields(struct ez_blobio *io, int depth);

static void tagz_dump_struct(struct ez_blobio *io, int depth) {
	uint8_t h = blobio_readb(io);
	uint32_t stag;
	//uint32_t scount = 1;

	char s[depth+1];
	memset(s, ' ', depth);
	s[depth] = 0;

	// must be a struct
	// what about a struct list? hmm
	if ((h & EZT_TYPE) != EZT_STRUCT)
		return;

	stag = blobio_readi(io, (h & EZT_TAGSIZE) >> EZT_TAGSHIFT);
	//if (h & EZT_COUNTSIZE)
	//	count = blob_readi(io, (h & EZT_COUNTSIZE) >> EZT_COUNTSHIFT);

	printf("%s[%02x] %d = {\n", s, h, stag);
	
	tagz_dump_fields(io, depth+4);
}

static void tagz_dump_fields(struct ez_blobio *io, int depth) {
	int h;
	
	char s[depth+1];
	memset(s, ' ', depth);
	s[depth] = 0;
		
	// read fields
	while ((h = blobio_readb(io)) != EZT_END) {
		uint32_t ftag;
		uint32_t fcount = 0;
		int sc = (h & EZT_DATASIZE) >> EZT_DATASHIFT;
		int use_count = 0;
		
		ftag = blobio_readi(io, (h & EZT_TAGSIZE) >> EZT_TAGSHIFT);
		if ((h & EZT_COUNTSIZE) != EZT_NOCOUNT) {
			fcount = blobio_readi(io, (h & EZT_COUNTSIZE) >> EZT_COUNTSHIFT);
			use_count = 1;
		}

		switch (h & EZT_TYPE) {
		case EZT_INT8:
		case EZT_INT16:
		case EZT_INT32:
		case EZT_INT64:
			if (use_count) {
				void *v = blobio_take(io, fcount * (1<<sc));
				if (v) {
					printf("%s[%02x] integer-%-2d %d [%d] = {\n", s, h, (1<<sc)*8, ftag, fcount);
					blobio_dumphex(v, fcount * (1<<sc), s);
					printf("%s}\n", s);
				}
			} else {
				printf("%s[%02x] integer-%-2d %d = %02lx\n", s, h, (1<<sc)*8, ftag, blobio_readi(io, sc));
			}
			break;
		case EZT_FLOAT32:
			if (use_count) {
				printf("%sfloat %d [%d] = {\n", s, ftag, fcount);
				for (int i=0;i<fcount;i++)
					printf("  %f\n", blobio_readf(io));
				printf("%s}\n", s);
			} else {
				printf("%sfloat %d = %f\n", s, ftag, blobio_readf(io));
			}
			break;
		case EZT_FLOAT64:
			if (use_count) {
				printf("%s%d double[%d] = {\n", s, ftag, fcount);
				for (int i=0;i<fcount;i++)
					printf("%s %f\n", s, blobio_readd(io));
				printf("%s}\n", s);
			} else {
				printf("%s%d double = %f\n", s, ftag, blobio_readd(io));
			}
			break;
		case EZT_STRUCT:
			if (use_count) {
				printf("%sstruct %d [%d] = {\n", s, ftag, fcount);
				for (int i=0;i<fcount;i++) {
					tagz_dump_struct(io, depth+4);
				}
				printf("%s}\n", s);
			} else {
				tagz_dump_fields(io, depth+4);
			}
			break;
		default:
			printf("%s[%02x] ??\n", s, h);
			break;
		}
		printf("%s[%02x] END $%04lx\n", s, h, io->index-1);
		printf("%s}\n", s);
	}
}

/**
 * ezt format is self-describing so can be dumped directly.
 */
void ez_tagz_dump(ez_blob *blob) {
	struct ez_blobio io = {
		.data = blob->data,
		.size = blob->size,
		.mode = BLOBIO_READ
	};

	tagz_dump_struct(&io, 4);
}
