/* ez-blob.h: Serialising description.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#ifndef _EZ_BLOB_H
#define _EZ_BLOB_H

#include <stddef.h>

/**
 * This is a C structure annotation for implementing a serialiser.
 *
 * It is table driven and can describe most common C data structures,
 * including nested and linked structures.
 *
 * A structure is described by an array of ez_blob_desc.  The first
 * entry must be of type EZ_BLOB_PK, it's 'offset' must be the size
 * of the structure, and .u.count must be the number of fields
 * following.
 * 
 * As an example:
 * 
 * typedef struct dbdisk {
 * 	int id;
 * 	char *uuid;
 * 	char *label;
 * 	char *type;
 * } dbdisk;
 * 
 * Has a table such as this, using the init macros.
 * 
 * static ez_blob_desc DISK_ENC[] = {
 * 	EZ_BLOB_START(dbdisk, 0, 3),
 * 	EZ_BLOB_STRING(dbdisk, 1, uuid),
 * 	EZ_BLOB_STRING(dbdisk, 2, label),
 * 	EZ_BLOB_STRING(dbdisk, 3, type),
 * };
 * 
 * The 'id' field and it's semantics are serialiser dependent.
 * 
 */
typedef struct ez_blob_desc {
	unsigned short bd_type;		// type of field
	unsigned short bd_id;		// used to tag encoded entries (currently unused)
	unsigned int bd_offset;		// offset into structure, if type == EZ_BLOB_PK then this is sizeof(struct)
	union {
		const struct ez_blob_desc *table;// embedded structure or pointer
		unsigned int length;
	} u;
#define bd_table u.table
#define bd_length u.length
} ez_blob_desc;

/**
 * Field type
 */
typedef enum ez_blob_desc_type {
	EZ_BLOB_INT8,
	EZ_BLOB_INT16,
	EZ_BLOB_INT32,
	EZ_BLOB_INT64,
	EZ_BLOB_FLOAT32,
	EZ_BLOB_FLOAT64,
	EZ_BLOB_STRING,			// normal c strring
	EZ_BLOB_BLOB,			// embedded ez_blob for arbitrary bytes.
	EZ_BLOB_STRUCT,			// bd_table points to another descriptor list, object embedded
	EZ_BLOB_POINTER,		// bd_table points to another descriptor list, object pointer
	EZ_BLOB_LIST,			// ez_list nodes.
	EZ_BLOB_LAST = EZ_BLOB_LIST,
	EZ_BLOB_PK = 16,	 	// MUST always be first element in descriptor and occur only once
	EZ_BLOB_NULLABLE = 0x8000,	// Pointer may be NULL for STRING, POINTER
	EZ_BLOB_TYPE = 0xff,		// Type mask
	EZ_BLOB_END
} ez_blob_desc_type;

/** 
 * This is compatible with lmdb MDB_val
 */
typedef struct ez_blob {
	size_t size;
	void *data;
} ez_blob;

#define EZ_BLOB_START(s, id, len) { EZ_BLOB_PK, 0, sizeof(s), .u.length = len }
#define EZ_BLOB_INT8(s, id, f) { EZ_BLOB_INT8, id, offsetof(s, f) }
#define EZ_BLOB_INT16(s, id, f) { EZ_BLOB_INT16, id, offsetof(s, f) }
#define EZ_BLOB_INT32(s, id, f) { EZ_BLOB_INT32, id, offsetof(s, f) }
#define EZ_BLOB_INT64(s, id, f) { EZ_BLOB_INT64, id, offsetof(s, f) }
#define EZ_BLOB_FLOAT32(s, id, f) { EZ_BLOB_FLOAT32, id, offsetof(s, f) }
#define EZ_BLOB_FLOAT64(s, id, f) { EZ_BLOB_FLOAT64, id, offsetof(s, f) }
#define EZ_BLOB_STRING(s, id, f) { EZ_BLOB_STRING, id, offsetof(s, f) }
#define EZ_BLOB_STRING_NULL(s, id, f) { EZ_BLOB_STRING | EZ_BLOB_NULLABLE, id, offsetof(s, f) }
#define EZ_BLOB_BLOB(s, id, f) { EZ_BLOB_BLOB, id, offsetof(s, f) }
#define EZ_BLOB_STRUCT(s, id, f, other) {  EZ_BLOB_STRUCT, id, offsetof(s, f), .u.table = other }
#define EZ_BLOB_POINTER(s, id, f, other) { EZ_BLOB_POINTER, id, offsetof(s, f), .u.table = other }
#define EZ_BLOB_POINTER_NULL(s, id, f, other) { EZ_BLOB_POINTER | EZ_BLOB_NULLABLE, id, offsetof(s, f), .u.table = other }
#define EZ_BLOB_LIST(s, id, f, other) { EZ_BLOB_LIST, id, offsetof(s, f), .u.table = other }

/**
 * Allocate a new structure to hold a blob.
 *
 * The structure is initialised by ez_blob_init().
 *
 * @param d the descriptor table for the blob.
 */
void *ez_blob_alloc(const ez_blob_desc *d);

/**
 * Initialise a struct to the default state.
 *
 * This sets all values to zero and lists to empty.
 *
 * @param d descriptor table matching struct.
 * @param p pointer to struct base.
 */
void ez_blob_init(const ez_blob_desc *d, void *p);

/**
 * Free struct unmarshalled by ez_blob_decode().
 *
 * @param d descriptor table matching struct.
 * @param p decoded struct from ez_blob_decode().
 */
void ez_blob_free(const ez_blob_desc *d, void *p);

/**
 * Free the contents of a structure, but don't free the container.
 *
 * Use ez_blob_init() if you want to re-use the object.
 *
 * @param p structure to free.  p will not be freed.
 */
void ez_blob_free_raw(const ez_blob_desc *d, void *p);

/**
 * Debug function to dump blob contents.
 */
void ez_blob_print(const ez_blob_desc *d, const void *a, int depth);

#endif
