/* ez-set.h: Low level hash table.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#ifndef _EZ_SET_H
#define _EZ_SET_H

#include <ez-node.h>

/**
 * A basic set / hash table implementation.
 *
 * All entries must start with a struct ez_entry and contain enough
 * information to caclulate the hash and equality.  Entries may only be
 * members of one set at a time.
 *
 * A general purpose hash table can be implemented with appropriate
 * initialisation and using an entry like the following:

    struct ez_bucket {
        struct ez_node bn;
	void *key;
	void *data;
    };

 *
 * For simplicity the table automatically rehashes up on put at a load
 * factor above 1.0 and down on remove at a load factor below 0.5.
 */

typedef struct ez_set ez_set;
typedef struct ez_set_scan ez_set_scan;

/**
 * Base set structure.
 */
struct ez_set {
	struct ez_node **table;	// table of chains
	int mask;		// size of the table - 1
	int count;		// number of entries in the set

	unsigned int (*hash)(const void *entry);
	int (*equals)(const void *a, const void *b);
	void (*free)(void *);
};

/**
 * State data for iterator.
 */
struct ez_set_scan {
	struct ez_set *set;
	struct ez_node *e;
	struct ez_node *p;
	int idx;
};

/**
 * Create a new set.
 *
 * @param h uninitialised set.
 * @param hash function which hashes entries.
 * @param equals function which compares entries for equality.
 * @param efree used to free nodes.  May be null.
 */
ez_set *ez_set_new(unsigned int (*hash)(const void *), int (*equals)(const void *, const void *), void (*efree)(void *));

/**
 * Free the set.
 *
 * The set will be cleared first.
 */
void ez_set_free(ez_set *h);

/**
 * Clear all entries and reset the internal size.
 *
 * If the set was initialised with a non-null free function, it will
 * be invoked on all nodes.
 */
void ez_set_clear(ez_set *h);

/**
 * Get the number of entries in the set.
 */
int ez_set_size(ez_set *h);

/**
 * Calculate the hash of a node.
 */
static __inline__ unsigned int ez_set_node_hash(ez_set *h, const void *e) {
	return h->hash(e);
}

/**
 * Compare nodes for equality.
 */
static __inline__ int ez_set_node_equals(ez_set *h, const void *a, const void *b) {
	return h->equals(a, b);
}

/**
 * Free a node.  If no node free function is set this is a nop.
 */
static __inline__ void ez_set_node_free(ez_set *h, void *e) {
	if (h->free)
		h->free(e);
}

/**
 * Put a table entry.
 *
 * @param h initialised set.
 * @param e entry to put.
 * @return the previous entry which equals e, if any.
 */
void *ez_set_put(ez_set *h, void *e);

/**
 * Get a table entry.
 *
 * @param h initialised set.
 * @param key key describing the item to retrieve.
 * @return the entry which equals e, if any.
 */
void *ez_set_get(ez_set *h, const void *key);

/**
 * Remove a table entry.
 *
 * @param h initialised set.
 * @param key key of entry to remove.
 * @return the entry removed, if any.
 */
void *ez_set_remove(ez_set *h, const void *key);

/**
 * Start an iterative scan of all set entries.
 *
 * The set must not be modified while the entries are being iterated
 * other than through ez_scan_remove().
 *
 * As with most hash tables, the iteration order is undefined.
 *
 * The scan can be abandoned at any time and @param scan does not need
 * to be disposed.
 *
 * @param h set to iterate.
 * @param scan uninitialised ez_set_scan structure.
 * @return the first node, or null if none exists.
 */
void *ez_set_scan_init(ez_set *h, ez_set_scan *scan);

/**
 * Retrieve the current node.
 *
 * This is just scan->e.
 */
static __inline__ void *ez_set_scan_get(ez_set_scan *scan) {
	return scan->e;
}

/**
 * Advance to the next node.
 *
 * Example to iterate all entries:

	ez_set *set = ...;
	ez_set_scan scan;
	struct entry_type *entry;

	for (entry = ez_scan_init(set, &scan); entry; entry = ez_scan_next(&scan)) {
		... work on entry ...
	}

 * @return the next node or NULL.
 */
void *ez_set_scan_next(ez_set_scan *scan);

/**
 * Remove the current node and advance to the next node.
 *
 * The node is not freed allowing it to be moved to other sets.
 *
 * Example to iterate and remove all entries:

	ez_set *set = ...;
	ez_set_scan scan;
	struct entry_type *next;

	for (next = ez_scan_init(set, &scan); next;) {
		struct entry_type *entry = next;

		next = ez_scan_remove(&scan);

		... work on entry ...
	}
 */
void *ez_set_scan_remove(ez_set_scan *scan);

/**
 * A basic string hash function.
 */
unsigned int ez_hash_string(const char *s);

/**
 * Simple array hash function.
 */
unsigned int ez_hash_array(const void *s, size_t size);

/**
 * 32-bit integer hash function.
 */
unsigned int ez_hash_int32(uint32_t v);

/**
 * 64-bit integer hash function.
 */
unsigned int ez_hash_int64(uint64_t v);

static unsigned int ez_set_node_hash(ez_set *h, const void *e) __attribute__((always_inline));
static int ez_set_node_equals(ez_set *h, const void *a, const void *b) __attribute__((always_inline));
static void ez_set_node_free(ez_set *h, void *e) __attribute__((always_inline));
static void *ez_set_scan_get(ez_set_scan *scan) __attribute__((always_inline));

#endif
