
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <assert.h>

#include "ez-blob.h"
#include "ez-list.h"

#include "ez-blob-blob.h"
#include "ez-blob-tagz.h"
#include "ez-blob-xdrn.h"

#include "ez-blob-io.h"

/*
  TODO: some bad data tests.

  This depends a lot on the structure and the format.
  
  * lengths outside of data
  * counts that will exhaust memory
  * field id's out of range
  * data corruption (wrong output)
  * meta-data corruption (wrong/invalid structure)
  * truncated data
  * too much data

  Primarily it shouldn't crash or corrupt memory.
  With explicit knowledge of format, it should fail on corrupt meta-data.
 */

struct simple {
	char b;
	short s;
	int a;
	int64_t l;
	float c;
	double d;
};

static const ez_blob_desc simple_DESC[] = {
	EZ_BLOB_START(struct simple, 0, 6),
	EZ_BLOB_INT8(struct simple, 1, b),
	EZ_BLOB_INT16(struct simple, 2, s),
	EZ_BLOB_INT32(struct simple, 3, a),
	EZ_BLOB_INT64(struct simple, 4, l),
	EZ_BLOB_FLOAT32(struct simple, 5, c),
	EZ_BLOB_FLOAT64(struct simple, 6, d),
};

struct arrays {
	uint32_t count;
	char *string;
	ez_blob array;
};

static const ez_blob_desc array_DESC[] = {
	EZ_BLOB_START(struct arrays, 0, 3),
	EZ_BLOB_INT32(struct arrays, 1, count),
	EZ_BLOB_STRING(struct arrays, 2, string),
	EZ_BLOB_BLOB(struct arrays, 3, array),
};

struct embedded {
	struct simple s;
	struct arrays a;
};

static const ez_blob_desc embed_DESC[] = {
	EZ_BLOB_START(struct embedded, 0, 2),
	EZ_BLOB_STRUCT(struct embedded, 1, s, simple_DESC),
	EZ_BLOB_STRUCT(struct embedded, 2, a, array_DESC),
};

struct tree {
	struct tree *left;
	struct tree *right;

	char *s;
};

static const ez_blob_desc tree_DESC[] = {
	EZ_BLOB_START(struct tree, 0, 3),
	EZ_BLOB_STRING(struct tree, 1, s),
	EZ_BLOB_POINTER_NULL(struct tree, 2, left, tree_DESC),
	EZ_BLOB_POINTER_NULL(struct tree, 3, right, tree_DESC),
};

struct listnode {
	ez_node ln;
	char *name;
	ez_blob value;
};

static const ez_blob_desc listnode_DESC[] = {
	EZ_BLOB_START(struct listnode, 0, 2),
	EZ_BLOB_STRING(struct listnode, 1, name),
	EZ_BLOB_BLOB(struct listnode, 2, value),
};

struct list {
	char *name;
	ez_list list;
};

static const ez_blob_desc list_DESC[] = {
	EZ_BLOB_START(struct list, 0, 2),
	EZ_BLOB_STRING(struct list, 1, name),
	EZ_BLOB_LIST(struct list, 2, list, listnode_DESC),
};

struct intsize {
	int16_t value16;
	int32_t value32;
	int64_t value64;
};

static const ez_blob_desc intsize_DESC[] = {
	EZ_BLOB_START(struct intsize, 7, 3),
	EZ_BLOB_INT16(struct intsize, 1, value16),
	EZ_BLOB_INT32(struct intsize, 2, value32),
	EZ_BLOB_INT64(struct intsize, 3, value64),
};

struct stringsize {
	char *value;
};

static const ez_blob_desc stringsize_DESC[] = {
	EZ_BLOB_START(struct stringsize, 8, 1),
	EZ_BLOB_STRING(struct stringsize, 1, value),
};

static void dumphex(const char *data, size_t size, const char *prefix) {
	blobio_dumphex(data, size, prefix);
}

static void blob_print(const ez_blob_desc *d, const void *a, int depth) {
	ez_blob_print(d, a, depth);
}

static int blob_equals(const ez_blob_desc *desc, const void *a, const void *b) {
	int e = 1;
	for (int i=0, dlen=desc->bd_length; e && i < dlen; i++) {
		const ez_blob_desc *d = &desc[i+1];
		const void *u = a + d->bd_offset;
		const void *v = b + d->bd_offset;

		switch (d->bd_type) {
		case EZ_BLOB_INT8:
			e &= (*(uint8_t *)u) == (*(uint8_t *)v);
			break;
		case EZ_BLOB_INT16:
			e &= (*(uint16_t *)u) == (*(uint16_t *)v);
			break;
		case EZ_BLOB_INT32:
			e &= (*(uint32_t *)u) == (*(uint32_t *)v);
			break;
		case EZ_BLOB_INT64:
			e &= (*(uint64_t *)u) == (*(uint64_t *)v);
			break;
		case EZ_BLOB_FLOAT32:
			e &= (*(float *)u) == (*(float *)v);
			break;
		case EZ_BLOB_FLOAT64:
			e &= (*(double *)u) == (*(double *)v);
			break;
		case EZ_BLOB_STRING | EZ_BLOB_NULLABLE:
			if (!*(char **)u || !*(char **)v) {
				e &= (*(char **)u) == (*(char **)v);
				break;
			}
			// falls through
		case EZ_BLOB_STRING:
			if (!*(char **)u || !*(char **)v) {
				printf("strings can't be null\n");
				e = 0;
			} else {
				e &= strcmp(*(char **)u, *(char **)v) == 0;
			}
			break;
		case EZ_BLOB_BLOB: {
			const struct ez_blob *ua = u;
			const struct ez_blob *va = v;

			e &= ua->size == va->size
				&& memcmp(ua->data, va->data, ua->size) == 0;
			break;
		}
		case EZ_BLOB_STRUCT:
			e &= blob_equals(d->bd_table, u, v);
			break;
		case EZ_BLOB_POINTER | EZ_BLOB_NULLABLE:
			if (!*(void **)u || !*(void **)v) {
				e &= *(void **)u == *(void **)v;
				break;
			}
			// falls through
		case EZ_BLOB_POINTER:
			e &= blob_equals(d->bd_table, *(void **)u, *(void **)v);
			break;
		case EZ_BLOB_LIST: {
			ez_list *ul = (void *)u;
			ez_list *vl = (void *)v;

			e &= ez_list_size(ul) == ez_list_size(vl);
			for (ez_node *uw = ez_list_head(ul), *un = ez_node_succ(uw),
				     *vw = ez_list_head(vl), *vn = ez_node_succ(uw);
			     un && vn;
			     uw = un, un = ez_node_succ(un),
				     vw = vn, vn = ez_node_succ(vn)) {
				e &= blob_equals(d->bd_table, uw, vw);
			}
			
			break;
		}
		default:
			printf("cmp unknown field type=%d id=%d\n", d->bd_type, d->bd_id);
			e = 0;
			break;
		}
		if (!e) {
			printf("cmp failed in struct %d at field %d type %04x\n", desc->bd_id, d->bd_id, d->bd_type);
			abort();
		}
	}

	return e;
}

struct test_funcs {
	const char *name;
	int (*blob_encode)(const ez_blob_desc *d, const void *p, ez_blob *blob);
	void *(*blob_decode)(const ez_blob_desc *desc, const ez_blob *blob);
	int (*blob_decode_raw)(const ez_blob_desc *desc, const ez_blob *blob, void *p);
};

static int doprint = 1;

static void test_basics(const ez_blob_desc *d, void *src, struct test_funcs *funcs) {
	ez_blob blob, blob2;
	void *t;
	size_t dsize = d->bd_offset;
	char tmp[dsize + 64] __attribute__ ((aligned(64)));
	int res;

	if (doprint) {
		printf("source\n");
		blob_print(d, src, 4);
	}
	
	res = funcs->blob_encode(d, src, &blob);
	assert(res == 0);
	if (doprint)
		dumphex(blob.data, blob.size, "serial: ");

	// check decode equals
	t = funcs->blob_decode(d, &blob);
	assert(t != NULL);

	if (doprint) {
		printf("decoded\n");
		blob_print(d, t, 4);
	}
	assert(blob_equals(d, t, src));

	// check encode-decoded equals source
	funcs->blob_encode(d, t, &blob2);
	assert(blob.size == blob2.size);
	assert(memcmp(blob.data, blob2.data, blob2.size) == 0);
	free(blob2.data);
	
	ez_blob_free(d, t);

	// check raw decode stays within bounds of struct
	memset(tmp, 0xbe, sizeof(tmp));
	funcs->blob_decode_raw(d, &blob, &tmp[32]);
	//assert(t != NULL);
	//assert(blob_equals(d, tmp+32, src));
	if (doprint)
		dumphex(tmp, sizeof(tmp), "object: ");

	for (int i=0;i<32;i++) {
		assert((tmp[i] & 0xff) == 0xbe);
		assert((tmp[i + 32 + dsize] & 0xff) == 0xbe);
	}
	ez_blob_free_raw(d, tmp+32);
	free(blob.data);
}

static void test_simple(struct test_funcs *funcs) {
	struct simple src = {
		'z', 0xff0f, 1, 0xf01040a04010dULL, 3.14159f, 5644941221133
	};
	const ez_blob_desc *d = simple_DESC;

	printf("test simple\n");
	test_basics(d, &src, funcs);
}

static void test_arrays(struct test_funcs *funcs) {
	char data[7] = { 1, 2, 3, 4, 5, 6, 7 };
	struct arrays src = {
		32,
		"Bob was here.",
		.array.size = 7,
		.array.data = data
	};
	const ez_blob_desc *d = array_DESC;

	printf("test arrays\n");
	test_basics(d, &src, funcs);
}

static void test_struct(struct test_funcs *funcs) {
	char data[10] = { 1, 2, 3, 4, 5, 6, 7, 1, 2, 3 };
	struct embedded src = {
		{
			-1, -1, -1, -1, 1337.4004, 6.022E23,
		},
		{
			55, "Jane was there.", .array.size = 10, .array.data = data
		}
	};
	const ez_blob_desc *d = embed_DESC;

	printf("test embedded structures\n");
	test_basics(d, &src, funcs);
}

static void test_rawtree(struct test_funcs *funcs) {
	struct tree src[6] = {
		{ &src[1], &src[3], "root" },
		{ NULL, &src[2], "left" },
		{ NULL, NULL, "left-right" },
		{ &src[4], &src[5], "right" },
		{ NULL, NULL, "right-left" },
		{ NULL, NULL, "right-right" },
	};
	const ez_blob_desc *d = tree_DESC;

	printf("test raw tree\n");
	test_basics(d, &src, funcs);
}

static void test_list(struct test_funcs *funcs) {
	struct list list = {
		.name = "a list",
		.list = EZ_INIT_LIST(list.list)
	};
	struct listnode nodes[3] = {
		{ .name = "node 0", .value.size = 5, .value.data = "data0" },
		{ .name = "node 1", .value.size = 5, .value.data = "data1" },
		{ .name = "node 2", .value.size = 5, .value.data = "data2" },
	};

	list.name = "a list";
	//	ez_list_init(&list.
	for (int i=0;i<3;i++)
		ez_list_addtail(&list.list, &nodes[i]);
	
	const ez_blob_desc *d = list_DESC;

	printf("test list\n");
	test_basics(d, &list, funcs);
}

static void test_intsize(struct test_funcs *funcs) {
	uint64_t sizes[8] = { 0x00, 0xff, 0x101, 0xffff, 0x10002, 0xffffffff, 0x100000003ULL, ~0ULL };
	const ez_blob_desc *d = intsize_DESC;
	struct intsize src;

	// Checks that the sizes reconstruct to their original values
	
	for (int i=0;i<8;i++) {
		uint64_t size = sizes[i];

		src.value16 = size;
		src.value32 = size;
		src.value64 = size;
		
		printf("test int value %016lx\n", size);
		test_basics(d, &src, funcs);
	}
}

static void test_stringsize(struct test_funcs *funcs) {
	int sizes[6] = { 0, 1, 255, 256, 65535, 65536 };
	const ez_blob_desc *d = stringsize_DESC;
	struct stringsize src;
	int sprint = doprint;
	
	for (int i=0;i<6;i++) {
		int size = sizes[i];
		src.value = malloc(size+1);
		memset(src.value, ' ', size);
		src.value[size] = 0;
		printf("test string size %d\n", size);
		doprint = size <= 256;
		test_basics(d, &src, funcs);
		free(src.value);
	}
	doprint = sprint;
}

// the corruption tests don't know the format details so can't tell what a given
// change should cause.
// however they should not crash, cause leaks, or valgrind oddness.

static void test_corrupt_bit(struct test_funcs *funcs) {
	// this is definitely not robust
	if (strcmp(funcs->name, "legacy") == 0) {
		printf("Not supposed to be safe\n");
		return;
	}
	
	char data[10] = { 1, 2, 3, 4, 5, 6, 7, 1, 2, 3 };
	struct embedded src = {
		{
			-1, -1, -1, -1, 1337.4004, 6.022E23,
		},
		{
			55, "Jane was there.", .array.size = 10, .array.data = data
		}
	};
	const ez_blob_desc *d = embed_DESC;
	struct embedded *dst;
	int res;
	ez_blob blob;

	// test 1 corrupt bit in every position
	
	// we don't know the internals so which ones would fail, but
	// can check for bad memory accesses with valgrind or crash
	
	int good = 0, bad = 0;
	res = funcs->blob_encode(d, &src, &blob);
	assert(res == 0);
	for (int i=0;i<blob.size;i++) {
		uint8_t *data = blob.data;
		for (int j=0;j<8;j++) {
			data[i] ^= (1<<j);
			dst = funcs->blob_decode(d, &blob);
			data[i] ^= (1<<j);
			
			if (dst) {
				good++;
				ez_blob_free(d, dst);
			} else {
				bad++;
			}
		}
	}
	free(blob.data);
	printf(" info: %d failed, %d decoded - but not important\n", bad, good);
}

static void test_corrupt_byte(struct test_funcs *funcs) {
	// this is definitely not robust
	if (strcmp(funcs->name, "legacy") == 0) {
		printf("Not supposed to be safe\n");
		return;
	}
	
	char data[10] = { 1, 2, 3, 4, 5, 6, 7, 1, 2, 3 };
	struct embedded src = {
		{
			-1, -1, -1, -1, 1337.4004, 6.022E23,
		},
		{
			55, "Jane was there.", .array.size = 10, .array.data = data
		}
	};
	const ez_blob_desc *d = embed_DESC;
	struct embedded *dst;
	int res;
	ez_blob blob;

	int good = 0, bad = 0;
	res = funcs->blob_encode(d, &src, &blob);
	assert(res == 0);
	for (int i=0;i<blob.size;i++) {
		uint8_t *data = blob.data;
		uint8_t save = data[i];

		for (int j=0;j<256;j++) {
			data[i] = j;
			dst = funcs->blob_decode(d, &blob);
			if (dst) {
				good++;
				ez_blob_free(d, dst);
			} else {
				bad++;
			}
		}
		data[i] = save;
	}
	free(blob.data);
	printf(" info: %d failed, %d decoded - but not important\n", bad, good);
}

struct test_funcs backends[] = {
	{
		"legacy",
		ez_blob_encode,
		ez_blob_decode,
		ez_blob_decode_raw
	},
	{
		"tagz",
		ez_tagz_encode,
		ez_tagz_decode,
		ez_tagz_decode_raw
	},
	{
		"xdrn",
		ez_xdrn_encode,
		ez_xdrn_decode,
		ez_xdrn_decode_raw
	},	
};

struct {
	const char *name;
	void (*test)(struct test_funcs *);
} tests[] = {
	{ "primitives", test_simple },
	{ "arrays", test_arrays },
	{ "structs", test_struct },
	{ "tree as links", test_rawtree },
	{ "lists", test_list },
	{ "string sizes", test_stringsize },
	{ "int sizes", test_intsize },
	{ "corrupted bit", test_corrupt_bit },
	{ "corrupted byte", test_corrupt_byte },
};

int main(int argc, char **argv) {
	int nfunc = sizeof(backends)/sizeof(backends[0]);
	int ntest = sizeof(tests)/sizeof(tests[0]);

	for (int i=0;i<nfunc;i++) {
		struct test_funcs *f = &backends[i];

		for (int j=0;j<ntest;j++) {
			printf("* **********************************************************************\n");
			printf("* test: %s / %s\n", f->name, tests[j].name);
			fflush(stdout);

			tests[j].test(f);
		}
	}
	return 0;
}
